--[[ TODO
- when saving pos do a "SetUserPlaced(nil)" to avoid scaling problems at load time.
- meta for name/icon/rank should be handled to account for proper rank.
- dontFade/vanish attribute to zap some fade effect
]]

local _, playerClass = UnitClass"player"
local spells = {}


local mt = { __index = function(t,k)
	if k == "hasGlyph" then -- on glyph change: if spell.glyphID then spell.hasGlyph = nil end
		for i=1, GetNumGlyphSockets() do
			if select( 3, GetGlyphSocketInfo(i) ) == t.glyphID then t[k] = true return true end
		end
		t[k] = false
		return false
	elseif k == "hasTalent" then -- on talent change: if spell.talentTree then spell.hasTalent = nil end
		local rank = select( 5, GetTalentInfo( t.talentTree, t.talentIndex ) )
		if rank > 0 then t[k], t.talentRank = true, rank return true end
		t[k] = false
		return false
	elseif k == "name" or k == "icon" or k == "rank" then
		t.name, t.rank, t.icon = GetSpellInfo( t.id ) -- UnitDebuff(unit,t.name,t.rank)
		return t[k]
	end
end }


local function AddSpells( ... )
	local spell
	for i=1, select( "#", ... ) do
		local obj, id = select( i, ... )
		if type(obj) == "table" then
			obj.id = id
			if not obj.glyphID then obj.hasGlyph = false end
			if not obj.talentTree then obj.hasTalent = false end
			spell = setmetatable( obj, mt )
		else
			spells[obj] = spell
		end
	end
end

if playerClass == "WARLOCK" then

	local CURSE = 1
	AddSpells(
--[[ Aftermath	     ]]	{ timer =   5 }, 18118,
--[[ Banish	     ]]	{ timer =  20, glyphID=56242, glyphAdd=5 }, 710,
			{ timer =  30, glyphID=56242, glyphAdd=5 }, 18647,
--[[ Corruption      ]]	{ timer =  12, fade=true }, 172,
			{ timer =  15, fade=true }, 6222,
			{ timer =  18, fade=true }, 6223, 7648, 11671, 11672, 25311, 27216, 47812, 47813,
--[[ Curse of Agony  ]]	{ timer =  24, fade=true, group=CURSE, glyphID=56241, glyphAdd=4 }, 980, 1014, 6217, 11711, 11712, 11713, 27218, 47863, 47864,
--[[ Curse of Doom   ]]	{ timer =  60, fade=true, group=CURSE }, 603, 30910, 47867,
--[[ Curse of Elem.  ]]	{ timer = 300, fade=true, dr=true, group=CURSE }, 1490, 11721, 11722, 27228, 47865,
--[[ Curse of Exhau. ]]	{ timer =  12, fade=true, group=CURSE }, 18223,
--[[ Curse of Reck.  ]]	{ timer = 120, fade=true, group=CURSE }, 704, 7658, 7659, 11717, 27226, 57595,
--[[ Curse of Tongue ]]	{ timer =  30, fade=true, dr=true, group=CURSE, pvpTimer=12 }, 1714, 11719, -- TOCHANGE
--[[ Curse of Weak.  ]]	{ timer = 120, fade=true, group=CURSE }, 702, 1108, 6205, 7646, 11707, 11708, 27224, 30909, 50511,
--[[ Enslave Demon   ]]	{ timer = 300 }, 1098, 11725, 11726,
--[[ Fear	     ]]	{ timer =  10, fade=true, dr=true }, 5782,
			{ timer =  15, fade=true, dr=true }, 6213,
			{ timer =  20, fade=true, dr=true }, 6215,
--[[ Haunt	     ]]	{ timer =  12, fade=true }, 48181, 59161, 59163, 59164,
--[[ Howl of Terror  ]]	{ timer =   6, fade=true, dr=true }, 5484,
			{ timer =   8, fade=true, dr=true }, 17928,
--[[ Immolate	     ]]	{ timer =  15, fade=true }, 348, 707, 1094, 2941, 11665, 11667, 11668, 25309, 27215, 47810, 47811,
--[[ Shadow Embrace  ]] { timer =  12 }, 32390, 32391, 32392, 32393, 32394,
--[[*Shadowburn      ]]	{ timer =   5 }, 29341,
--[[ Shadowfury	     ]] { timer =   3, dr=true }, 30283, 30413, 30414, 47846, 47847,
--[[ Succubus: Seduce]]	{ timer =  15, fade=true, dr=true, talentTree=2, talentIndex=8, talentMul=0.1 }, 6358,
--[[ Unstable Affli. ]]	{ timer =  15, fade=true }, 30108, 30404, 30405, 47841, 47843
	)

elseif playerClass == "DRUID" then

	AddSpells(
--[[ Animal Soothe ]]	{ timer = 15 }, 2908, 8955, 9901, 26995,
--[[ Bash ]]		{ timer =  2, dr=true, talentTree=2, talentIndex=13, talentAdd=.5 }, 5211,
			{ timer =  3, dr=true, talentTree=2, talentIndex=13, talentAdd=.5 }, 6798,
			{ timer =  4, dr=true, talentTree=2, talentIndex=13, talentAdd=.5 }, 8983,
--[[ Cyclone ]]		{ timer =  6, dr=true }, 33786, -- Cyclone
--[[ Demoraliz.Roar ]]	{ timer = 30, fade=true }, 99, 1735, 9490, 9747, 9898, 26998,
--[[ Faerie Fire ]]	{ timer = 40, fade=true }, 770, 778, 9749, 9907, 26993, 48476, --[[Feral]] 16857, 17390, 17391, 17392, 27011, 48475,
--[[ Feral Charge ]]	{ timer =  4, dr=true }, 16979, -- immobilize effect
--[[ Hibernate ]]	{ timer = 20, fade=true, dr=true }, 2637,
			{ timer = 30, fade=true, dr=true }, 18657,
			{ timer = 40, fade=true, dr=true }, 18658,
--[[ Insect Swarm ]]	{ timer = 12, fade=true, talentTree=1, talentIndex=8, talentAdd=2 }, 5570, 24974, 24975, 24976, 24977, 27013, 48468,
--[[ Lacerate ]]	{ timer = 15, fade=true }, 33745, 48567, 48568,
--[[ Mangle ]]		{ timer = 12, fade=true, glyphID=54813, glyphAdd=6 }, 33878, --[[Bear]] 33986, 33987, 48563, 48564, --[[Cat]] 33876, 33982, 33983, 48565, 48566,
--[[ Moonfire ]]	{ timer =  9, fade=true, talentTree=1, talentIndex=8, talentAdd=3 }, 8921,
			{ timer = 12, fade=true, talentTree=1, talentIndex=8, talentAdd=3 }, 8924, 8925, 8926, 8927, 8928, 8929, 9833, 9834, 9835, 26987, 26988, 48462, 48463,
--[[ Pounce (stun) ]]	{ timer =  4, dr=true, talentTree=2, talentIndex=13, talentAdd=.5 }, 9005, 9823, 9827, 27006, 49803, -- Pounce (stun part) + talent
--[[ Pounce (bleed) ]]	{ timer = 18, fade=true }, 9007, 9824, 9826, 27007, 49804, -- Pounce Bleed (dot part)
--[[ Rake ]]		{ timer =  9, fade=true }, 1822, 1823, 1824, 9904, 27003,
--[[ Rip ]]		{ timer = 12, fade=true, glyphID=54818, glyphAdd=4 }, 1079, 9492, 9493, 9752, 9894, 9896, 27008, 49799, 49800,
--[[ Roots ]]		{ timer = 12, dr=true }, 339,
			{ timer = 15, dr=true }, 1062,
			{ timer = 18, dr=true }, 5195,
			{ timer = 21, dr=true }, 5196,
			{ timer = 24, dr=true }, 9852,
			{ timer = 27, dr=true }, 9853, 26989, 53308,
--[[ Starfire ]]	{ timer =  3 }, 16922 -- Starfire Stun (EFFECT)
		-- TODO: Natural Perfection,
	)

elseif playerClass == "DEATHKNIGHT" then

	AddSpells(
--[[ Chains of Ice ]]	{ timer = 10 }, 47805,
--[[ Death Grip		{ timer =  3 }, 49560, --]]
--[[ Ebon Plague ]]--	{ timer = 12, fade=true, talentTree=3, talentIndex=4, talentAdd=3 }, ,
--[[ Frost Fever ]]	{ timer = 15, fade=true, talentTree=3, talentIndex=4, talentAdd=3 }, 55095, -- Frost Fever (Icy Touch) ]]
--[[ Hungering Cold  ]]	{ timer = 10 }, 51209,
--[[ Chillblains ]]	{ timer = 10 }, 50434, 50435, 50436,
--[[*Mind Freeze     ]]	{ timer =  4, SPELL_INTERRUPT=true }, 47528, -- Mind Freeze (SPELL_INTERRUPT)
--[[ Plague Strike   ]]	{ timer = 15, fade=true, talentTree=3, talentIndex=4, talentAdd=3 }, 55078, 57601, 58840, 58844, -- Blood Plague (Plague Strike dot)
--[[ Strangulate     ]]	{ timer =  5, dr=true }, 47476, 49913, 49914, 49915, 49916,
--[[ Gnaw ]]		{ timer =  3, dr=true }, 47481,
--[[ Rune of Razorice ]]{ timer = 20 }, 51714,
--[[ Glyph.Heart.Strk ]]{ timer = 10 }, 58617
	)

elseif playerClass == "PRIEST" then

	AddSpells(
--[[ Blackout ]]	{ timer =  3, dr=true }, 15269,
--[[ Devouring Plague ]]{ timer = 24, fade=true }, 2944, 19276, 19277, 19278, 19279, 19280, 25467, 48299, 48300,
--[[ Divine Hymn ]]--	{ timer = 20, fade=true }, 47951,
--[[ Holy Fire ]]	{ timer =  7, fade=true }, 14914, 15262, 15263, 15264, 15265, 15266, 15267, 15261, 25384, 48134, 48135,
--[[ Mind Control ]]	{ timer = 60, dr=true }, 605,
--[[ Mind Soothe ]]	{ timer = 15 }, 453,
--[[ Psychic Screem ]]	{ timer =  8, fade=true, dr=true, glyphID=55676, glyphAdd=1 }, 8122, 8124, 10888, 10890,
--[[ Shackle Undead ]]	{ timer = 30, dr=true }, 9484,
			{ timer = 40, dr=true }, 9485,
			{ timer = 50, dr=true }, 10955,
--[[ Silence ]]		{ timer =  5, dr=true }, 15487,
--[[ SW: Pain ]]	{ timer = 18, fade=true }, 589, 594, 970, 992, 2767, 10892, 10893, 10894, 25367, 25368, 48124, 48125,
--[[ Vampiric Embrace ]]{ timer = 60 }, 15286,
--[[ Vampiric Touch ]]	{ timer = 15, fade=true }, 34914, 34916, 34917, 48159, 48160
	)

elseif playerClass == "MAGE" then

	AddSpells(
--[[ Blast Wave  ]]	{ timer =  6 }, 11113, 13018, 13019, 13020, 13021, 27133, 33933, 42944, 42945,
--[[ Cone of Cold ]]	{ timer =  8 }, 120, 8492, 10159, 10160, 10161, 27087, 42930, 42931,
--[[ Counterspell ]]	{ timer =  8, dr=true, SPELL_INTERRUPT=true }, 2139,
--[[ Deep Freeze ]]	{ timer =  5, dr=true }, 44572,
--[[ Dragon's Breath ]]	{ timer =  5 }, 31661, 33041, 33042, 33042, 42949, 42950,
--[[ Fingers of Frost ]]{ timer = 15 }, 44544,
--[[ Fireball DoT ]]	{ timer =  4, fade=true, glyphID=56368, glyphAdd=-30 }, 133,
			{ timer =  6, fade=true, glyphID=56368, glyphAdd=-30 }, 143, 145,
			{ timer =  8, fade=true, glyphID=56368, glyphAdd=-30 }, 3140, 8400, 8401, 8402, 10148, 10149, 10150, 10151, 25306, 27070, 38692, 42832, 42833,
--[[ Flamestrike ]]	{ timer =  8, fade=true }, 2120, 2121, 8422, 8423, 10215, 10216, 27086, 42925, 42926,
--[[ Frost Armor ]]	{ timer =  5, group=1 }, 6136,
--[[ Frost Nova ]]	{ timer =  8, dr=true }, 122, 865, 6131, 10230, 27088, 42917,
--[[ Frostbite ]]	{ timer =  5, dr=true }, 12494, -- DR ?
--[[ Frostbolt effect]]	{ timer =  5, group=1, talentTree=3, talentIndex=7, talentAdd=1 }, 116,
			{ timer =  6, group=1, talentTree=3, talentIndex=7, talentAdd=1 }, 205, 837,
			{ timer =  7, group=1, talentTree=3, talentIndex=7, talentAdd=1 }, 7322, 8406,
			{ timer =  8, group=1, talentTree=3, talentIndex=7, talentAdd=1 }, 8407, 8408,
			{ timer =  9, group=1, talentTree=3, talentIndex=7, talentAdd=1 }, 10179, 10180, 10181, 25304, 27071, 27072, 38697, 42841, 42842,
--[[ Impact ]]		{ timer =  2 }, 12355,
--[[ Imp. CS ]]		{ timer =  0, talentTree=1, talentIndex=12, talentAdd=2 }, 18469,
--[[ Imp. Scorch ]]	{ timer = 30, fade=true }, 22959,
--[[ Living Bomb ]]	{ timer = 12, fade=true }, 44457, 55359, 55360,
--[[ Polymorph ]]	{ timer = 20, fade=true, dr=true }, 118,
			{ timer = 30, fade=true, dr=true }, 12824,
			{ timer = 40, fade=true, dr=true }, 12825,
			{ timer = 50, fade=true, dr=true }, 12826, 28271, 28272, 61025, 61305,
--[[ Slow ]]		{ timer = 15, fade=true, dr=true }, 31589, -- pvp timer != pve timer ?
--[[ Winter's Chill ]]	{ timer = 15 }, 12579
	)

else return end

-- CONST
local refreshRate = 0.100 -- sec
local BORDER_GAP = 6
local offsetTimer = 28 -- espacement horizontal entre timers
local offsetHeader = 18 -- espacement vertical entre le nom de la cible et les dots
local offsetTarget = offsetHeader + 50 -- espacement vertical des targets
local backdrop = {
	bgFile = "Interface/Buttons/WHITE8X8",
	edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
	tile = true, tileSize = 16, edgeSize = 12,
	insets = { left = 2, right = 2, top = 2, bottom = 2 } }

-- DATA
local frame, OnUpdate = CreateFrame( "Frame", "Ara_Dotimer_Frame" )
local targets = {}
local nbTargets = 0
local playerGUID
local targetGUID, targetLevel, targeted, prevTargetGUID, prevTargetLevel
local focusGUID,  focusLevel,  focused
local config
local defaultConfig = { timersGrowRight=true, scale=1, x=330, y=400 }

local border = {
	bgFile = "Interface/Buttons/WHITE8X8",
	edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
	tile = true, tileSize = 16, edgeSize = 12,
	insets = { left = 2, right = 2, top = 2, bottom = 2 } }

local targetBorder = CreateFrame("Frame", "Ara_Dotimer_TargetBorder", frame)
targetBorder:SetFrameStrata"BACKGROUND"
targetBorder:Show()
targetBorder:SetAlpha(0)
targetBorder:SetBackdrop( border )
targetBorder:SetBackdropBorderColor( .2, .2, 1, .15 )
targetBorder:SetBackdropColor( .2,.2,1,.15 )

local focusBorder = CreateFrame("Frame", "Ara_Dotimer_FocusBorder", frame)
focusBorder:SetFrameStrata"BACKGROUND"
focusBorder:Show()
focusBorder:SetAlpha(0)
focusBorder:SetBackdrop( border )
focusBorder:SetBackdropBorderColor( 1, .2, .2, .2 )
focusBorder:SetBackdropColor( 1,.2,.2,.2 )


-- LOCAL GLOBALS
local   bit_or,  bit_and,  tremove, next, floor, format, wipe, UnitDebuff, GetTime, UnitGUID, GameFontNormal, GameFontHighlightSmall, GameFontHighlight, NumberFontNormalLarge =
	bit.bor, bit.band, tremove, next, floor, format, wipe, UnitDebuff, GetTime, UnitGUID, GameFontNormal, GameFontHighlightSmall, GameFontHighlight, NumberFontNormalLarge

-- POOLS
local tables = {} -- table reuse

local function AcquireTable( ... )
	local table = tremove(tables) or {}
	for i = 1, select( "#", ... ), 2 do
		local key, value = select( i, ... )
		table[key] = value
	end
	return table
end

local function ReleaseTable(table)
	tables[#tables+1] = wipe(table)
end


local fontstrings = {} -- fontstring reuse

local function AcquireFS(doseCounter)
	local fs = tremove(fontstrings)
	if not fs then
		fs = frame:CreateFontString( nil, "BACKGROUND" )
		fs:SetJustifyH("CENTER")
		fs.hidden = true
	end
	fs:SetFontObject( doseCounter and NumberFontNormalLarge or GameFontHighlightSmall )
	return fs
end

local function ReleaseFS(fs)
	fs:ClearAllPoints()
	fs:Hide()
	fs.hidden = true
	fontstrings[#fontstrings+1] = fs
end


local headers = {}

local function AcquireHeader()
	local header = tremove(headers)
	if not header then
		header = frame:CreateFontString( nil, "BACKGROUND" )
		header:SetJustifyH("LEFT")
		header:SetFontObject(GameFontNormal)
		header.hidden = true
	end
	return header
end

local function ReleaseHeader(header)
	header:Hide()
	header.hidden = true
	headers[#headers+1] = header
end


local function ClearTargetHL()
	targeted.header:SetFontObject(GameFontNormal)
	targetBorder:SetAlpha(0)
	targeted.targeted = nil
	targeted = nil
end

local function SetTargetHL(target)
	if targeted then ClearTargetHL() end
	targeted = target
	if not target then return end
	target.header:SetFontObject(GameFontHighlight)
	targetBorder:SetAlpha(1)
	targetBorder:SetWidth( target.header:GetWidth()+BORDER_GAP*2 )
	targetBorder:SetHeight( target.header:GetHeight() + offsetHeader + 16 + BORDER_GAP*2 )
	targetBorder:SetPoint( "TOPLEFT", frame, config.timersGrowRight and "TOPLEFT" or "TOPRIGHT", config.timersGrowRight and -BORDER_GAP or -targetBorder:GetWidth()+BORDER_GAP, -offsetTarget*target.pos +BORDER_GAP )
	target.targeted = true
end


local function ClearFocusHL()
	focusBorder:SetAlpha(0)
	focused.focused = nil
	focused = nil
end

local function SetFocusHL(focus)
	if focused then ClearFocusHL() end
	focused = focus
	if not focus then return end
	focusBorder:SetAlpha(1)
	focusBorder:SetWidth( focus.header:GetWidth()+BORDER_GAP*2)
	focusBorder:SetHeight( focus.header:GetHeight() + offsetHeader + 16 + BORDER_GAP*2 )
	focusBorder:SetPoint( "TOPLEFT", frame, config.timersGrowRight and "TOPLEFT" or "TOPRIGHT", config.timersGrowRight and -BORDER_GAP or -focusBorder:GetWidth()+BORDER_GAP, -offsetTarget*focus.pos +BORDER_GAP )
--	focusBorder:SetWidth( focus.header:GetWidth()+BORDER_GAP*2)
--	focusBorder:SetHeight( focus.header:GetHeight() + offsetHeader + 16 + BORDER_GAP*2 )
	focus.focused = true
end


local function CleanUITarget( target, timers )
	local removedPos, guid = target.pos, target.guid
	ReleaseHeader( target.header )
	ReleaseTable( timers )
	if target.targeted then ClearTargetHL() end
	if target.focused then ClearFocusHL() end
	ReleaseTable( target )
	targets[guid] = nil
	nbTargets = nbTargets - 1
	-- remove OnUpdate if no target
	if nbTargets == 0 then
		return frame:SetScript("OnUpdate",nil)
	end
	-- reorder ui targets
	for k,v in next, targets do
		if v.pos > removedPos then
			v.pos = v.pos - 1
		end
	end
end

local function RemoveTimer( target, spell, keepTarget )
	local timers = target.timers
	local timer = timers[spell]
	if timer then
		local removedPos = timer.pos
		ReleaseFS( timer.fs )
		if timer.dose then ReleaseFS( timer.dose ) end
		ReleaseTable( timer )
		timers[spell] = nil
		if target.nbTimers == 1 and not keepTarget then
			return CleanUITarget( target, timers )
		end
		--if removedPos < nbTimers then -- reorder ui timers
		for k,v in next, timers do
			if v.pos > removedPos then
				v.pos = v.pos - 1
			end
		end
		target.nbTimers = target.nbTimers - 1
	end
end

local function AddTimer( guid, name, spell, timestamp, offset )
	local target = targets[guid]
	if not target then
		target = AcquireTable( "guid", guid, "name", name, "timers", AcquireTable(),
			"isPlayer", strsub( guid, 3, 5 ) == "000", "pos", nbTargets, "nbTimers", 0, "header", AcquireHeader() )
		target.level =  targetGUID==guid and targetLevel or prevTargetGUID==guid and prevTargetLevel
			or focusGUID==guid and focusLevel or UnitGUID"mouseover"==guid and UnitLevel"mouseover" or 0
		if target.level==0 then
			target.level="??"
		elseif target.level==-1 then
			target.level = "BOSS"
		end
		targets[guid] = target
		nbTargets = nbTargets + 1
		if nbTargets == 1 then
			frame:SetScript("OnUpdate",OnUpdate)
		end
	elseif spell.group then -- curses & co
		for sp in next, target.timers do
			if sp.group == spell.group then
				RemoveTimer( target, sp, true )
				break
			end
		end
	end

	local timers = target.timers
	local timer = timers[spell]
	if not timer then -- NEW TIMER
		timer = AcquireTable( "name", spell.name, "fs", AcquireFS() )
		timers[spell] = timer
		target.nbTimers = target.nbTimers + 1
	end

	local spellTimer, duration = offset or spell.timer
	if not offset then
		if spell.dr and target.isPlayer then
			local unit = targetGUID==guid and "target" or focusGUID==guid and "focus" or UnitGUID"mouseover"==guid and "mouseover"
			if unit then
				duration = select(6, UnitDebuff( unit, spell.name, spell.rank, "PLAYER,PET" ) )
				if duration then spellTimer = duration end
			end
		end
		if not duration then
			if spell.hasGlyph then spellTimer = spellTimer + spell.glyphAdd end
			if spell.hasTalent then
				if spell.talentAdd then
					spellTimer = spellTimer + spell.talentRank * spell.talentAdd
				else	spellTimer = spellTimer * ( 1 + spell.talentRank * spell.talentMul ) end
			end
		end
	end
	timer.startTime, timer.expireTime = timestamp, timestamp + spellTimer

	timer.fs:SetAlpha(1) -- ADDED fade out
	timer.fadeout = nil -- ADDED fade out

	-- sort timers ( expire: 300 -> 0 )
	local insertedPos = timer.pos or target.nbTimers
	timer.pos = insertedPos
	for k, v in next, timers do
		if timer.expireTime > v.expireTime then
			if insertedPos > v.pos then
				timer.pos = timer.pos - 1
				v.pos = v.pos + 1
			end
		elseif timer.expireTime < v.expireTime then
			if insertedPos < v.pos then
				timer.pos = timer.pos + 1
				v.pos = v.pos - 1
			end
		end
	end
end

local function RemoveTarget( guid )
	local target = targets[guid]
	if target then
		local timers = target.timers
		for k,v in next, timers do
			ReleaseFS( v.fs )
			if v.dose then ReleaseFS( v.dose ) end
			ReleaseTable( v )
			timers[k] = nil
		end
		return CleanUITarget( target, timers )
	end
end


OnUpdate = function( self, elapsed )
	self.timer = self.timer - elapsed
	if self.timer >= 0 then return else self.timer = refreshRate end
	local currentTime = GetTime()
	-- process each target
	for guid, target in next, targets do
		local timers = target.timers
		local targetPosChanged = target.pos ~= target.prevPos
		-- repositionne le nom de la target si necessaire
		if targetPosChanged then
			target.prevPos = target.pos
			local header = target.header
			-- affiche si cach
			if header.hidden then
				header:Show()
				header:SetFormattedText( "[%s] %s", target.level ~= 0 and tostring(target.level) or "??", target.name or "unknown" )
				if targetGUID==guid then SetTargetHL(target) end
				if focusGUID==guid then SetFocusHL(target) end
				header.hidden = false
			end
			if config.timersGrowRight then
				header:SetPoint( "TOPLEFT", frame, "TOPLEFT", 0, -offsetTarget * target.pos )
			else
				header:SetPoint( "TOPLEFT", frame, "TOPRIGHT", -header:GetStringWidth(), -offsetTarget * target.pos )
			end
			if target.targeted then -- repositionne le target highlight
				targetBorder:SetPoint( "TOPLEFT", frame, config.timersGrowRight and "TOPLEFT" or "TOPRIGHT", config.timersGrowRight and -BORDER_GAP or -targetBorder:GetWidth()+BORDER_GAP, -offsetTarget*target.pos +BORDER_GAP )
			end
			if target.focused then -- repositionne le focus highlight
				focusBorder:SetPoint( "TOPLEFT", frame, config.timersGrowRight and "TOPLEFT" or "TOPRIGHT", config.timersGrowRight and -BORDER_GAP or -focusBorder:GetWidth()+BORDER_GAP, -offsetTarget*target.pos +BORDER_GAP )
			end
		end
		-- process each timer
		for spell, timer in next, timers do
			local timeLeft = timer.expireTime - currentTime
			if timeLeft > -15 then
				if timeLeft >= 5 then -- less granular
					timeLeft = floor( timeLeft )
					if timer.prevTime ~= timeLeft then
						timer.prevTime = timeLeft
						local ftime = timeLeft < 60 and format( "%is", timeLeft) or format( "%i:%.2i", floor( timeLeft / 60 ), floor( timeLeft ) % 60 )
						timer.fs:SetFormattedText( "\124T%s:25:25:0:0\124t\n\124cff%s%s\124r", spell.icon, timeLeft < 10 and "ffff20" or "20ff20", ftime )
					end
				elseif timeLeft >= 0 then
					timer.fs:SetFormattedText( "\124T%s:25:25:0:0\124t\n\124cffff2020%.1f\124r", spell.icon, timeLeft )
				elseif not timer.fadeout then
					if not spell.fade then return RemoveTimer( target, spell ) end
					timer.fadeout = true
					if timer.dose then ReleaseFS( timer.dose ) timer.dose = nil end
					timer.fs:SetAlpha(.6)
				--	timer.fs:SetFormattedText( "\124T%s:25:25:0:0\124t\n ", spell.icon ) -- REM, see below
				end
				if timer.fadeout then -- ADDED timer for 0 --> -15 sec
					timer.fs:SetFormattedText( "\124T%s:25:25:0:0\124t\n\124cff800080%i\124r", spell.icon, timeLeft )
				end
				-- repositionne le timer si necessaire, l'affiche si cach.
				if targetPosChanged or timer.prevPos ~= timer.pos then
					local fs = timer.fs
					if config.timersGrowRight then
						fs:SetPoint( "TOPLEFT", frame, "TOPLEFT", offsetTimer * (timer.pos-1), -offsetTarget * target.pos - offsetHeader )
					else
						fs:SetPoint( "TOPLEFT", frame, "TOPRIGHT", -offsetTimer * (timer.pos), -offsetTarget * target.pos - offsetHeader )
					end
					timer.prevPos = timer.pos
					if fs.hidden then
						fs:Show()
						fs.hidden = false
					end
				end
			else	 -- fixes some issues
				RemoveTimer( target, spell )
			end
		end
	end
end

local combatEvents = {
	["SPELL_AURA_APPLIED"] = true,
	["SPELL_AURA_REFRESH"] = true,
	["SPELL_AURA_REMOVED"] = true,
	["SPELL_INTERRUPT"] = true,
	["SPELL_AURA_APPLIED_DOSE"] = true,
	["SPELL_AURA_REMOVED_DOSE"] = true,
	["UNIT_DIED"] = true,
	["UNIT_DESTROYED"] = true,
}

local function OnCombatEvent( self, event, timestamp, combatEvent, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ... )
	if not combatEvents[combatEvent] then return end
	if srcGUID == playerGUID or srcFlags == 0x1111 --[[ pet ]] then
		-- target must be unfriendly
		if bit_and( dstFlags, COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL ) == 0 then return end
		local spell = spells[...]
		if not spell then return end
		if combatEvent == "SPELL_AURA_APPLIED" or combatEvent == "SPELL_AURA_REFRESH" then
			return AddTimer( dstGUID, dstName, spell, GetTime() ) --timestamp )
		elseif combatEvent == "SPELL_AURA_REMOVED" then
			local target = targets[dstGUID]
			if not target then return end
			local timer = target.timers[ spell ]
			if not timer then return end
			if spell.fade then timer.expireTime = GetTime() - 0.001 else RemoveTimer(target, spell) end
		elseif combatEvent == "SPELL_INTERRUPT" then -- TODO: merge with SPELL_AURA_APPLIED ?
			AddTimer( dstGUID, dstName, spell, GetTime() )
		else--if combatEvent == "SPELL_AURA_APPLIED_DOSE" or combatEvent == "SPELL_AURA_REMOVED_DOSE" then
			local target = targets[dstGUID]
			if not target then return end
			local timer = target.timers[spell]
			if not timer then return end

			AddTimer( dstGUID, dstName, spell, GetTime() )
			if not timer.dose then
				timer.dose = AcquireFS(true)
				timer.dose:SetPoint( "TOP", timer.fs, "TOP", 0, -2 )
				timer.dose:SetTextColor( 1, 1, 1, .75 )
				timer.dose:Show()
			end
			return timer.dose:SetText( select(5,...) )
		end
	elseif combatEvent == "UNIT_DIED" or combatEvent == "UNIT_DESTROYED" then -- ou "else" tout court ?
		return RemoveTarget( dstGUID )
	end
end


local function OnEvent( self, event, arg1, arg2, ... )
	if event == "PLAYER_TARGET_CHANGED" then
		prevTargetGUID, prevTargetLevel = targetGUID,		targetLevel
		targetGUID,	targetLevel	= UnitGUID"target",	UnitLevel"target"
		if nbTargets == 0 then return end
		local target = targets[targetGUID]
		if targeted and targeted ~= target or target then return SetTargetHL(target) end
	elseif event == "PLAYER_FOCUS_CHANGED" then
		focusGUID = UnitGUID"focus"
		focusLevel = UnitLevel"focus"
		if nbTargets == 0 then return end
		local focus = targets[focusGUID]
		if focused and focused ~= focus or focus then return SetFocusHL(focus) end
	elseif event == "PLAYER_REGEN_ENABLED" then
		-- fin du combat, on vire les mobs qui ont reset
		for guid in next, targets do
			if strsub(guid,5,5) == "3" then -- mob
				RemoveTarget( guid )
			end
		end
	elseif event == "PLAYER_DEAD" then
		-- si joueur mort, on retire toutes les targets
		for guid in next, targets do
			RemoveTarget( guid )
		end
	elseif event == "PLAYER_TALENT_UPDATE" then
		for id, spell in next, spells do
			if spell.talentTree then spell.hasTalent = nil end
		end
	elseif event == "ADDON_LOADED" then
		if arg1 ~= "Ara_Dotimer" then return else self:UnregisterEvent(event) end
		AraDotimerDB = AraDotimerDB or defaultConfig
		config = AraDotimerDB
		for k, v in next, defaultConfig do
			if config[k] == nil then config[k] = v end
		end
		frame:SetScale( config.scale )
		frame:SetPoint( "CENTER", UIParent, "BOTTOMLEFT", config.x, config.y )
	elseif event == "PLAYER_LOGIN" then
		playerGUID = UnitGUID"player"
	else -- GLYPH
		for id, spell in next, spells do
			if spell.glyphID then spell.hasGlyph = nil end
		end
	end
end


frame:SetScript("OnEvent",OnEvent)
frame:RegisterEvent"PLAYER_TARGET_CHANGED"
frame:RegisterEvent"PLAYER_FOCUS_CHANGED"
frame:RegisterEvent"PLAYER_REGEN_ENABLED"
frame:RegisterEvent"PLAYER_DEAD"
frame:RegisterEvent"PLAYER_TALENT_UPDATE"
frame:RegisterEvent"GLYPH_ADDED" frame:RegisterEvent"GLYPH_REMOVED" frame:RegisterEvent"GLYPH_UPDATED"
frame:RegisterEvent"ADDON_LOADED"

frame.timer = refreshRate
frame:SetWidth(140)
frame:SetHeight(60)
frame:SetMovable(true)

local extraFrame = CreateFrame"Frame"
extraFrame:SetScript("OnEvent",OnCombatEvent)
extraFrame:RegisterEvent"COMBAT_LOG_EVENT_UNFILTERED"

local function ToggleAnchor()
	local show = not frame:IsMouseEnabled()
	frame:SetBackdrop( show and backdrop or nil )
	frame:EnableMouse( show )
	if show then frame:SetBackdropColor( .3, 0, .3, .5 ) end
end

frame:SetClampedToScreen(true) -- ADDED
frame:EnableMouseWheel(true) -- ADDED
frame:SetScript("OnMouseWheel", function(self, delta)
	local scale = self:GetScale() + delta * .05
	config.scale = scale
	self:SetScale(scale)
end)

SLASH_ARADOTIMER1, SLASH_ARADOTIMER2 = "/dt", "/dot"
SlashCmdList["ARADOTIMER"] = function( param, self )
	param = strlower(param)
	if param == "grow" then
		config.timersGrowRight = not config.timersGrowRight
		return print( "[|cffffb366Ara|r|cffffff00Dotimer|r] Timers grow direction set to |cff40ff40" .. (config.timersGrowRight and "RIGHT" or "LEFT") .. "|r." )
	end
	ToggleAnchor()
end

frame:SetScript( "OnMouseDown", function(self,arg) if arg~="LeftButton" then self:StopMovingOrSizing();ToggleAnchor() else self:StartMoving() end end )
frame:SetScript( "OnMouseUp", function(self,arg)
	self:StopMovingOrSizing()
	config.x, config.y = self:GetCenter()
	self:SetUserPlaced(nil)
end )

return IsLoggedIn() and OnEvent(frame, "PLAYER_LOGIN") or frame:RegisterEvent"PLAYER_LOGIN"